import { CommandHandler, ICommandHandler } from '@nestjs/cqrs';
import { CreateAdCommand } from './create-ad.command';
import { Inject } from '@nestjs/common';
import { AD_REPOSITORY, AD_ROUTE_PROVIDER } from '@modules/ad/ad.di-tokens';
import { AdEntity } from '@modules/ad/core/domain/ad.entity';
import { AdRepositoryPort } from '../../ports/ad.repository.port';
import { AggregateID, ConflictException } from '@mobicoop/ddd-library';
import { AdAlreadyExistsException } from '@modules/ad/core/domain/ad.errors';
import { RouteProviderPort } from '../../ports/route-provider.port';
import { Role } from '@modules/ad/core/domain/ad.types';
import {
  Path,
  PathCreator,
  PathType,
  TypedRoute,
} from '@modules/ad/core/domain/path-creator.service';
import { Point } from '../../types/point.type';
import { Waypoint } from '../../types/waypoint.type';
import { Waypoint as WaypointValueObject } from '@modules/ad/core/domain/value-objects/waypoint.value-object';

@CommandHandler(CreateAdCommand)
export class CreateAdService implements ICommandHandler {
  constructor(
    @Inject(AD_REPOSITORY)
    private readonly repository: AdRepositoryPort,
    @Inject(AD_ROUTE_PROVIDER)
    private readonly routeProvider: RouteProviderPort,
  ) {}

  async execute(command: CreateAdCommand): Promise<AggregateID> {
    const roles: Role[] = [];
    if (command.driver) roles.push(Role.DRIVER);
    if (command.passenger) roles.push(Role.PASSENGER);
    const pathCreator: PathCreator = new PathCreator(
      roles,
      command.waypoints.map(
        (waypoint: Waypoint) =>
          new WaypointValueObject({
            position: waypoint.position,
            lon: waypoint.lon,
            lat: waypoint.lat,
          }),
      ),
    );
    let typedRoutes: TypedRoute[];
    try {
      typedRoutes = await Promise.all(
        pathCreator.getBasePaths().map(async (path: Path) => ({
          type: path.type,
          route: await this.routeProvider.getBasic(path.waypoints),
        })),
      );
    } catch (e: any) {
      throw new Error('Unable to find a route for given waypoints');
    }

    let driverDistance: number | undefined;
    let driverDuration: number | undefined;
    let passengerDistance: number | undefined;
    let passengerDuration: number | undefined;
    let points: Point[] | undefined;
    let fwdAzimuth: number | undefined;
    let backAzimuth: number | undefined;
    typedRoutes.forEach((typedRoute: TypedRoute) => {
      if (typedRoute.type !== PathType.PASSENGER) {
        driverDistance = typedRoute.route.distance;
        driverDuration = typedRoute.route.duration;
        points = typedRoute.route.points;
        fwdAzimuth = typedRoute.route.fwdAzimuth;
        backAzimuth = typedRoute.route.backAzimuth;
      }
      if (typedRoute.type !== PathType.DRIVER) {
        passengerDistance = typedRoute.route.distance;
        passengerDuration = typedRoute.route.duration;
        if (!points) points = typedRoute.route.points;
        if (!fwdAzimuth) fwdAzimuth = typedRoute.route.fwdAzimuth;
        if (!backAzimuth) backAzimuth = typedRoute.route.backAzimuth;
      }
    });
    if (points && fwdAzimuth && backAzimuth) {
      const ad = AdEntity.create({
        id: command.id,
        driver: command.driver,
        passenger: command.passenger,
        frequency: command.frequency,
        fromDate: command.fromDate,
        toDate: command.toDate,
        schedule: command.schedule,
        seatsProposed: command.seatsProposed,
        seatsRequested: command.seatsRequested,
        strict: command.strict,
        waypoints: command.waypoints,
        points,
        driverDistance,
        driverDuration,
        passengerDistance,
        passengerDuration,
        fwdAzimuth,
        backAzimuth,
      });
      try {
        await this.repository.insertExtra(ad, 'ad');
        return ad.id;
      } catch (error: any) {
        if (error instanceof ConflictException) {
          throw new AdAlreadyExistsException(error);
        }
        throw error;
      }
    }
    throw new Error('Route error');
  }
}
