import { Frequency, Role } from '@modules/ad/core/domain/ad.types';
import { Selector } from '../algorithm.abstract';
import { ScheduleItem } from '../match.query';
import { Waypoint } from '../../../types/waypoint.type';
import { Point } from '../../../types/point.type';
import { CandidateEntity } from '@modules/ad/core/domain/candidate.entity';
import { AdEntity } from '@modules/ad/core/domain/ad.entity';

export class PassengerOrientedSelector extends Selector {
  select = async (): Promise<CandidateEntity[]> => {
    const queryStringRoles: QueryStringRole[] = [];
    if (this.query.driver)
      queryStringRoles.push({
        query: this._createQueryString(Role.DRIVER),
        role: Role.DRIVER,
      });
    if (this.query.passenger)
      queryStringRoles.push({
        query: this._createQueryString(Role.PASSENGER),
        role: Role.PASSENGER,
      });

    return (
      await Promise.all(
        queryStringRoles.map<Promise<AdsRole>>(
          async (queryStringRole: QueryStringRole) =>
            <AdsRole>{
              ads: await this.repository.getCandidateAds(queryStringRole.query),
              role: queryStringRole.role,
            },
        ),
      )
    )
      .map((adsRole: AdsRole) =>
        adsRole.ads.map((adEntity: AdEntity) =>
          CandidateEntity.create({
            id: adEntity.id,
            role: adsRole.role,
            waypoints: adEntity.getProps().waypoints,
          }),
        ),
      )
      .flat();
  };

  private _createQueryString = (role: Role): string =>
    [
      this._createSelect(role),
      this._createFrom(),
      'WHERE',
      this._createWhere(role),
    ]
      .join(' ')
      .replace(/\s+/g, ' '); // remove duplicate spaces for easy debug !

  private _createSelect = (role: Role): string =>
    [
      `SELECT \
        ad.uuid,driver,passenger,frequency,public.st_astext(ad.waypoints) as waypoints,\
        "fromDate","toDate",\
        "seatsProposed","seatsRequested",\
        strict,\
        "fwdAzimuth","backAzimuth",\
        ad."createdAt",ad."updatedAt",\
        si.uuid as "scheduleItemUuid",si.day,si.time,si.margin,si."createdAt" as "scheduleItemCreatedAt",si."updatedAt" as "scheduleItemUpdatedAt"`,
      role == Role.DRIVER ? this._selectAsDriver() : this._selectAsPassenger(),
    ].join();

  private _selectAsDriver = (): string =>
    `${this.query.driverRoute?.duration} as duration,${this.query.driverRoute?.distance} as distance`;

  private _selectAsPassenger = (): string =>
    `"driverDuration" as duration,"driverDistance" as distance`;

  private _createFrom = (): string =>
    'FROM ad LEFT JOIN schedule_item si ON ad.uuid = si."adUuid"';

  private _createWhere = (role: Role): string =>
    [
      this._whereRole(role),
      this._whereStrict(),
      this._whereDate(),
      this._whereSchedule(role),
      this._whereAzimuth(),
      this._whereProportion(role),
      this._whereRemoteness(role),
    ]
      .filter((where: string) => where != '')
      .join(' AND ');

  private _whereRole = (role: Role): string =>
    role == Role.PASSENGER ? 'driver=True' : 'passenger=True';

  private _whereStrict = (): string =>
    this.query.strict
      ? this.query.frequency == Frequency.PUNCTUAL
        ? `frequency='${Frequency.PUNCTUAL}'`
        : `frequency='${Frequency.RECURRENT}'`
      : '';

  private _whereDate = (): string =>
    `(\
      (\
        "fromDate" <= '${this.query.fromDate}' AND "fromDate" <= '${this.query.toDate}' AND\
        "toDate" >= '${this.query.toDate}' AND "toDate" >= '${this.query.fromDate}'\
      ) OR (\
        "fromDate" >= '${this.query.fromDate}' AND "fromDate" <= '${this.query.toDate}' AND\
        "toDate" <= '${this.query.toDate}' AND "toDate" >= '${this.query.fromDate}'\
      ) OR (\
        "fromDate" <= '${this.query.fromDate}' AND "fromDate" <= '${this.query.toDate}' AND\
        "toDate" <= '${this.query.toDate}' AND "toDate" >= '${this.query.fromDate}'\
      ) OR (\
        "fromDate" >= '${this.query.fromDate}' AND "fromDate" <= '${this.query.toDate}' AND\
        "toDate" >= '${this.query.toDate}' AND "toDate" >= '${this.query.fromDate}'\
      )\
    )`;

  private _whereSchedule = (role: Role): string => {
    const schedule: string[] = [];
    // we need full dates to compare times, because margins can lead to compare on previous or next day
    // -first we establish a base calendar (up to a week)
    const scheduleDates: Date[] = this._datesBetweenBoundaries(
      this.query.fromDate,
      this.query.toDate,
    );
    // - then we compare each resulting day of the schedule with each day of calendar,
    //  adding / removing margin depending on the role
    scheduleDates.map((date: Date) => {
      this.query.schedule
        .filter(
          (scheduleItem: ScheduleItem) => date.getDay() == scheduleItem.day,
        )
        .map((scheduleItem: ScheduleItem) => {
          switch (role) {
            case Role.PASSENGER:
              schedule.push(this._wherePassengerSchedule(date, scheduleItem));
              break;
            case Role.DRIVER:
              schedule.push(this._whereDriverSchedule(date, scheduleItem));
              break;
          }
        });
    });
    if (schedule.length > 0) {
      return ['(', schedule.join(' OR '), ')'].join('');
    }
    return '';
  };

  private _wherePassengerSchedule = (
    date: Date,
    scheduleItem: ScheduleItem,
  ): string => {
    let maxDepartureDatetime: Date = new Date(date);
    maxDepartureDatetime.setHours(parseInt(scheduleItem.time.split(':')[0]));
    maxDepartureDatetime.setMinutes(parseInt(scheduleItem.time.split(':')[1]));
    maxDepartureDatetime = this._addMargin(
      maxDepartureDatetime,
      scheduleItem.margin as number,
    );
    // we want the min departure time of the driver to be before the max departure time of the passenger
    return `make_timestamp(\
                ${maxDepartureDatetime.getFullYear()},\
                ${maxDepartureDatetime.getMonth() + 1},\
                ${maxDepartureDatetime.getDate()},\
                CAST(EXTRACT(hour from time) as integer),\
                CAST(EXTRACT(minute from time) as integer),0) - interval '1 second' * margin <=\
                make_timestamp(\
                  ${maxDepartureDatetime.getFullYear()},\
                  ${maxDepartureDatetime.getMonth() + 1},\
                  ${maxDepartureDatetime.getDate()},${maxDepartureDatetime.getHours()},${maxDepartureDatetime.getMinutes()},0)`;
  };

  private _whereDriverSchedule = (
    date: Date,
    scheduleItem: ScheduleItem,
  ): string => {
    let minDepartureDatetime: Date = new Date(date);
    minDepartureDatetime.setHours(parseInt(scheduleItem.time.split(':')[0]));
    minDepartureDatetime.setMinutes(parseInt(scheduleItem.time.split(':')[1]));
    minDepartureDatetime = this._addMargin(
      minDepartureDatetime,
      -(scheduleItem.margin as number),
    );
    // we want the max departure time of the passenger to be after the min departure time of the driver
    return `make_timestamp(\
                ${minDepartureDatetime.getFullYear()},
                ${minDepartureDatetime.getMonth() + 1},
                ${minDepartureDatetime.getDate()},\
                CAST(EXTRACT(hour from time) as integer),\
                CAST(EXTRACT(minute from time) as integer),0) + interval '1 second' * margin >=\
                make_timestamp(\
                  ${minDepartureDatetime.getFullYear()},
                  ${minDepartureDatetime.getMonth() + 1},
                  ${minDepartureDatetime.getDate()},${minDepartureDatetime.getHours()},${minDepartureDatetime.getMinutes()},0)`;
  };

  private _whereAzimuth = (): string => {
    if (!this.query.useAzimuth) return '';
    const { minAzimuth, maxAzimuth } = this._azimuthRange(
      this.query.backAzimuth as number,
      this.query.azimuthMargin as number,
    );
    if (minAzimuth <= maxAzimuth)
      return `("fwdAzimuth" <= ${minAzimuth} OR "fwdAzimuth" >= ${maxAzimuth})`;
    return `("fwdAzimuth" <= ${minAzimuth} AND "fwdAzimuth" >= ${maxAzimuth})`;
  };

  private _whereProportion = (role: Role): string => {
    if (!this.query.useProportion) return '';
    switch (role) {
      case Role.PASSENGER:
        return `(${this.query.passengerRoute?.distance}>(${this.query.proportion}*"driverDistance"))`;
      case Role.DRIVER:
        return `("passengerDistance">(${this.query.proportion}*${this.query.driverRoute?.distance}))`;
    }
  };

  private _whereRemoteness = (role: Role): string => {
    this.query.waypoints.sort(
      (firstWaypoint: Waypoint, secondWaypoint: Waypoint) =>
        firstWaypoint.position - secondWaypoint.position,
    );
    switch (role) {
      case Role.PASSENGER:
        return `\
        public.st_distance('POINT(${this.query.waypoints[0].lon} ${
          this.query.waypoints[0].lat
        })'::public.geography,direction)<\
        ${this.query.remoteness} AND \
        public.st_distance('POINT(${
          this.query.waypoints[this.query.waypoints.length - 1].lon
        } ${
          this.query.waypoints[this.query.waypoints.length - 1].lat
        })'::public.geography,direction)<\
        ${this.query.remoteness}`;
      case Role.DRIVER:
        const lineStringPoints: string[] = [];
        this.query.driverRoute?.points.forEach((point: Point) =>
          lineStringPoints.push(
            `public.st_makepoint(${point.lon},${point.lat})`,
          ),
        );
        const lineString = [
          'public.st_makeline( ARRAY[ ',
          lineStringPoints.join(','),
          '] )::public.geography',
        ].join('');
        return `\
        public.st_distance( public.st_startpoint(waypoints::public.geometry), ${lineString})<\
        ${this.query.remoteness} AND \
        public.st_distance( public.st_endpoint(waypoints::public.geometry), ${lineString})<\
        ${this.query.remoteness}`;
    }
  };

  private _datesBetweenBoundaries = (
    firstDate: string,
    lastDate: string,
    max = 7,
  ): Date[] => {
    const fromDate: Date = new Date(firstDate);
    const toDate: Date = new Date(lastDate);
    const dates: Date[] = [];
    let count = 0;
    for (
      let date = fromDate;
      date <= toDate;
      date.setDate(date.getDate() + 1)
    ) {
      dates.push(new Date(date));
      count++;
      if (count == max) break;
    }
    return dates;
  };

  private _addMargin = (date: Date, marginInSeconds: number): Date => {
    date.setTime(date.getTime() + marginInSeconds * 1000);
    return date;
  };

  private _azimuthRange = (
    azimuth: number,
    margin: number,
  ): { minAzimuth: number; maxAzimuth: number } => ({
    minAzimuth:
      azimuth - margin < 0 ? azimuth - margin + 360 : azimuth - margin,
    maxAzimuth:
      azimuth + margin > 360 ? azimuth + margin - 360 : azimuth + margin,
  });
}

export type QueryStringRole = {
  query: string;
  role: Role;
};

type AdsRole = {
  ads: AdEntity[];
  role: Role;
};
