import { Role } from '@modules/ad/core/domain/ad.types';
import {
  Path,
  PathCreator,
  PathType,
} from '@modules/ad/core/domain/path-creator.service';
import { Waypoint } from '@modules/ad/core/domain/value-objects/waypoint.value-object';

const originWaypoint: Waypoint = new Waypoint({
  position: 0,
  lat: 48.689445,
  lon: 6.17651,
});
const destinationWaypoint: Waypoint = new Waypoint({
  position: 1,
  lat: 48.8566,
  lon: 2.3522,
});
const intermediateWaypoint: Waypoint = new Waypoint({
  position: 1,
  lat: 48.74488,
  lon: 4.8972,
});
const destinationWaypointWithIntermediateWaypoint: Waypoint = new Waypoint({
  position: 2,
  lat: 48.8566,
  lon: 2.3522,
});

describe('Path Creator Service', () => {
  it('should create a path for a driver only', () => {
    const pathCreator: PathCreator = new PathCreator(
      [Role.DRIVER],
      [originWaypoint, destinationWaypoint],
    );
    const paths: Path[] = pathCreator.getBasePaths();
    expect(paths).toHaveLength(1);
    expect(paths[0].type).toBe(PathType.DRIVER);
  });
  it('should create a path for a passenger only', () => {
    const pathCreator: PathCreator = new PathCreator(
      [Role.PASSENGER],
      [originWaypoint, destinationWaypoint],
    );
    const paths: Path[] = pathCreator.getBasePaths();
    expect(paths).toHaveLength(1);
    expect(paths[0].type).toBe(PathType.PASSENGER);
  });
  it('should create a single path for a driver and passenger', () => {
    const pathCreator: PathCreator = new PathCreator(
      [Role.DRIVER, Role.PASSENGER],
      [originWaypoint, destinationWaypoint],
    );
    const paths: Path[] = pathCreator.getBasePaths();
    expect(paths).toHaveLength(1);
    expect(paths[0].type).toBe(PathType.GENERIC);
  });
  it('should create two different paths for a driver and passenger with intermediate waypoint', () => {
    const pathCreator: PathCreator = new PathCreator(
      [Role.DRIVER, Role.PASSENGER],
      [
        originWaypoint,
        intermediateWaypoint,
        destinationWaypointWithIntermediateWaypoint,
      ],
    );
    const paths: Path[] = pathCreator.getBasePaths();
    expect(paths).toHaveLength(2);
    expect(
      paths.filter((path: Path) => path.type == PathType.DRIVER),
    ).toHaveLength(1);
    expect(
      paths.filter((path: Path) => path.type == PathType.DRIVER)[0].waypoints,
    ).toHaveLength(3);
    expect(
      paths.filter((path: Path) => path.type == PathType.PASSENGER),
    ).toHaveLength(1);
    expect(
      paths.filter((path: Path) => path.type == PathType.PASSENGER)[0]
        .waypoints,
    ).toHaveLength(2);
  });
});
